import 'dart:convert';

import '../php_serializer.dart';

/// Parses a [String] which could be provided by Php via its function
/// `serialize()` and returns the resulting object.
///
/// If more complex objects should be deserialized, further details for those
/// have to be provided via a [List] of [PhpSerializationObjectInformation]
/// as the second argument.
dynamic phpDeserialize(String serializedString,
        {List<PhpSerializationObjectInformation>? knownClasses,
        NoMatchingObjectDeserializationInformation?
            fallbackObjectDeserialization}) =>
    _Deserializer(
            knownClasses ?? [],
            fallbackObjectDeserialization ??
                const ThrowExceptionOnMissingDeserializationInformation())
        .parse(serializedString);

/// Implement this interface to customize the action when there is not matching
/// information for deserialization.
/// {@category Fallback Behavior}
abstract class NoMatchingObjectDeserializationInformation {
  /// Return a [PhpSerializationObjectInformation] for the class identified by
  /// [classIdentifier]
  PhpSerializationObjectInformation handleDeserialization(
      String classIdentifier);
}

/// Will be thrown in case [phpDeserialize] fails for some reason
abstract class DeserializationException implements Exception {
  final String message;

  const DeserializationException(this.message);

  @override
  String toString() => message;
}

///Deserialization of an object failed because there was no matching
///[PhpSerializationObjectInformation] provided
class ObjectWithoutDeserializationInformationFound
    extends DeserializationException {
  const ObjectWithoutDeserializationInformationFound(String fqcn)
      : super(
            'An object with classname $fqcn couldn\'t be deserialized, since no deserialiazion-information was provided!');
}

///Deserialization of an object failed because the user-defined converter
///function threw an exception.
class CustomDeserializationFailed implements DeserializationException {
  final dynamic innerException;
  final Type objectType;

  @override
  String get message =>
      'An exception of type ${innerException.runtimeType.toString()} was thrown when trying to deserialize an Object of Type ${objectType.toString()}\n'
      'Inner ${innerException.toString()}';

  const CustomDeserializationFailed(this.objectType, this.innerException);
}

///Deserialization failed because the passed string was invalid
class InvalidSerializedString extends DeserializationException {
  const InvalidSerializedString(String serializedString)
      : super('Invalid serialized string: $serializedString');
}

class _Deserializer {
  final List<PhpSerializationObjectInformation> knownClasses;
  final NoMatchingObjectDeserializationInformation
      fallbackObjectDeserialization;

  const _Deserializer(this.knownClasses, this.fallbackObjectDeserialization);

  dynamic parse(String rawInput) {
    final repr = _StringRepresentation(
        rawInput, knownClasses, fallbackObjectDeserialization);
    return _parse(repr);
  }

  dynamic _parse(_StringRepresentation repr) {
    final identifier = repr.read(1, skip: 1);

    switch (identifier) {
      case 's':
        return _parseString(repr);
      case 'i':
        return _parseInt(repr);
      case 'd':
        return _parseDouble(repr);
      case 'a':
        return _parseArray(repr);
      case 'O':
        return _parseObject(repr);
      case 'N':
        return null;
      case 'b':
        return (repr.read(2) == '1;');
      default:
        throw InvalidSerializedString(repr.readAll());
    }
  }

  String _parseString(_StringRepresentation repr) {
    final lengthOfString = int.parse(repr.readUntil(delimiter: 58, skip: 2));
    final returnValue = repr.read(lengthOfString, skip: 2);
    return returnValue;
  }

  int _parseInt(_StringRepresentation repr) {
    final returnValue =
        repr.readUntil(delimiter: 59, skip: 1); //delimiter = ';'
    return int.parse(returnValue);
  }

  double _parseDouble(_StringRepresentation repr) {
    final returnValue =
        repr.readUntil(delimiter: 59, skip: 1); //delimiter = ';'
    return double.parse(returnValue);
  }

  dynamic _parseArray(_StringRepresentation repr,
      {bool allowSimplification = true}) {
    final arrayLength =
        int.parse(repr.readUntil(delimiter: 58, skip: 2)); //delimiter = ':'
    final possibleReturnValue = <dynamic, dynamic>{};
    var canBeSimplified = allowSimplification;

    for (var i = 0; i != arrayLength; ++i) {
      final key = _parse(repr);
      if (canBeSimplified && (key is! int || key != i)) {
        canBeSimplified = false;
      }
      final value = _parse(repr);
      possibleReturnValue[key] = value;
    }
    repr.skip(); //}

    if (canBeSimplified) {
      return List<dynamic>.generate(
          possibleReturnValue.length, (index) => possibleReturnValue[index]);
    }

    return possibleReturnValue;
  }

  dynamic _parseObject(_StringRepresentation repr) {
    final lengthOfName =
        int.parse(repr.readUntil(delimiter: 58, skip: 2)); //delimiter = ':'
    final classIdentifier = repr.read(lengthOfName, skip: 2);
    final parameterArray = Map<String, dynamic>.from(
        _parseArray(repr, allowSimplification: false));
    final objectInfo = repr.getObjectInformation(classIdentifier);
    final objectGenerator = (objectInfo.objectGenerator ??
        fallbackObjectDeserialization
            .handleDeserialization(classIdentifier)
            .objectGenerator);

    if (objectGenerator == null) {
      throw ObjectWithoutDeserializationInformationFound(classIdentifier);
    }

    try {
      return objectGenerator(parameterArray);
    } catch (e) {
      throw CustomDeserializationFailed(objectInfo.typeOf, e);
    }
  }
}

class _StringRepresentation {
  final List<int> _rawByteArray;
  int _offset = 0;
  final List<PhpSerializationObjectInformation> _knownClasses;
  final NoMatchingObjectDeserializationInformation
      fallbackObjectDeserialization;

  _StringRepresentation(
      String rawString, this._knownClasses, this.fallbackObjectDeserialization)
      : _rawByteArray = utf8.encode(rawString);

  void skip([int length = 1]) {
    _offset += length;
  }

  String read(int length, {int skip = 0}) {
    assert(!isDone);
    if (length > remaining) length = remaining;

    final returnValue = _rawByteArray.getRange(_offset, _offset + length);

    _offset += length;
    this.skip(skip);

    return utf8.decode(returnValue.toList(growable: false));
  }

  //: = 58; 0x3a
  //; = 59; 0x3b
  String readUntil({int delimiter = 58, int skip = 0}) {
    assert(!isDone);
    final offset = _offset;
    _offset = _rawByteArray.indexOf(delimiter, _offset);
    if (_offset == -1) {
      _offset = _rawByteArray.length;
    }

    final returnValue = utf8.decode(
        _rawByteArray.getRange(offset, _offset).toList(growable: false));

    this.skip(skip);
    return returnValue;
  }

  String readAll() => utf8.decode(_rawByteArray
      .getRange(_offset, _rawByteArray.length)
      .toList(growable: false));

  bool get isDone => _offset == _rawByteArray.length;

  int get remaining => _rawByteArray.length - _offset;

  PhpSerializationObjectInformation getObjectInformation(String identifier) {
    return _knownClasses.firstWhere(
        (element) => element.serializedClassName == identifier,
        orElse: () =>
            fallbackObjectDeserialization.handleDeserialization(identifier));
  }
}
